"""Main module."""
__version__ = '0.0.1'

from abc import ABC, abstractmethod
import logging

from hamcrest import anything, match_equality, equal_to, has_item, starts_with, \
    greater_than, greater_than_or_equal_to, less_than, less_than_or_equal_to, \
    close_to, contains_string, string_contains_in_order, equal_to_ignoring_case, \
    equal_to_ignoring_whitespace, not_none, none, any_of, all_of, is_not, \
    has_entry, has_entries

# TODO: did not utilize the "is_" function from hamcrest


from enum import Enum


class MatcherType(Enum):
    NOTHING = 'Nothing'
    ANYTHING = 'Anything'
    EQUAL_TO = 'EqualTo'
    STARTS_WITH = 'StartsWith'
    CONTAINS_STRING = 'ContainsString'
    CONTAINS_STRING_IN_ORDER = 'ContainsStringInOrder'
    EQUAL_TO_IGNORE_CASE = 'EqualToIgnoreCase'
    EQUAL_TO_IGNORE_WHITESPACE = 'EqualToIgnoreWhitespace'
    GREATER_THAN = 'GreaterThan'
    GREATER_THAN_EQUAL_TO = 'GreaterThanEqualTo'
    LESS_THAN = 'LessThan'
    LESS_THAN_EQUAL_TO = 'LessThanEqualTo'
    CLOSE_TO = 'CloseTo'
    NONE = 'None'
    NONE_OR_EMPTY = 'NoneOrEmpty'
    NOT_NONE = 'NotNone'
    NOT_NONE_OR_EMPTY = 'NotNoneOrEmpty'
    HAS_ENTRY = 'HasEntry'
    HAS_ENTRIES = 'HasEntries'


# pull values from list...support for list of input parameters and *list syntax
def pull_val(x):
    return x


class Matcher(ABC):

    def __init__(self, match_col_key):
        self.matcher_type = MatcherType.NOTHING
        self.match_col_key = match_col_key
        self.my_logger = logging.getLogger('matching')

    @abstractmethod
    def is_match(self, match_values, data_record) -> bool:
        pass

    def validate_key_exists(self, data_record) -> bool:
        """ Validate match-key-column exists in data record"""

        if self.match_col_key not in data_record:
            self.my_logger.warning((f"'{self.match_col_key}' not present"
                                    f" in data record \n\n{data_record}. Matcher will return False\n\n"))
            return False
        else:
            return True

    def get_key_val(self):
        """ Generate a value suitable for hashing, dictionary key"""
        return frozenset((self.matcher_type.value, self.match_col_key))

    def __repr__(self):
        return (f'{self.__class__.__name__}('
                f'{self.matcher_type!r}, {self.match_col_key!r})')

    def __str__(self):
        return (f'Matcher for {self.matcher_type.value!r} '
                f'matching on field {self.match_col_key!r}')

    def __eq__(self, other):
        if other.__class__ is self.__class__:
            return (self.matcher_type, self.match_col_key) == \
                (other.matcher_type, other.match_col_key)
        else:
            return NotImplemented

    def __hash__(self):
        return hash((self.__class__, self.matcher_type, self.match_col_key))


class NothingMatcher(Matcher):
    """
    Matcher never successfully matches any input. Always returns False.
    """

    def __init__(self, match_col_key):
        super().__init__(match_col_key)
        self.match_col_key = match_col_key
        self.matcher_type = MatcherType.NOTHING

    def is_match(self, match_values, data_record) -> bool:
        self.my_logger.info("No Matcher set, defaults to Nothing Matcher. Always False.")
        return False


class AnythingMatcher(Matcher):
    """
    Matcher always successfully matches any input. Always returns True.
    """

    def __init__(self, match_col_key):
        super().__init__(match_col_key)
        self.match_col_key = match_col_key
        self.matcher_type = MatcherType.ANYTHING
        self.my_matcher = anything(f"Anything for {match_col_key}")

    def is_match(self, match_values, data_record) -> bool:
        return match_equality(self.my_matcher) == data_record


class EqualTo(Matcher):
    """ Equal To matching style. Cast everything to str. """

    def __init__(self, match_col_key, ):
        super().__init__(match_col_key, )
        self.match_col_key = match_col_key
        self.matcher_type = MatcherType.EQUAL_TO

    def is_match(self, match_values, data_record) -> bool:
        if not self.validate_key_exists(data_record):
            return False

        if len(match_values) == 0:
            self.my_logger.warning("No Match Values provided, raising Error")
            raise NotImplementedError("Cannot use Equal To to check for empty "
                                      "string. Use None or Not_None.")

        elif isinstance(match_values, list):
            if len(match_values) == 0:
                # covered by len ==0 above
                pass

            elif len(match_values) == 1 and not isinstance(data_record[self.match_col_key], list):
                q_match_values = pull_val(*match_values)
                return (match_equality(equal_to(q_match_values))
                        == data_record[self.match_col_key])

            else:
                # has_item will iterate a sequence ...
                if isinstance(data_record[self.match_col_key], list):
                    check_list = []
                    for dr in data_record[self.match_col_key]:
                        check_list.append(match_equality(
                            has_item(equal_to(dr))) == [x for x in match_values])
                    return any(check_list)
                else:
                    return (match_equality(
                        has_item(equal_to(data_record[self.match_col_key]))) == [x for x in match_values])
        else:
            return (match_equality(equal_to(match_values))
                    == data_record[self.match_col_key])


class TextComparer(Matcher):

    def __init__(self, match_col_key, matcher_type):
        super().__init__(match_col_key)
        self.match_col_key = match_col_key

        if matcher_type == MatcherType.STARTS_WITH:
            self.matcher_type = MatcherType.STARTS_WITH
            self.my_matcher = starts_with

        elif matcher_type == MatcherType.CONTAINS_STRING:
            self.matcher_type = MatcherType.CONTAINS_STRING
            self.my_matcher = contains_string

        elif matcher_type == MatcherType.CONTAINS_STRING_IN_ORDER:
            self.matcher_type = MatcherType.CONTAINS_STRING_IN_ORDER
            self.my_matcher = string_contains_in_order

        elif matcher_type == MatcherType.EQUAL_TO_IGNORE_CASE:
            self.matcher_type = MatcherType.EQUAL_TO_IGNORE_CASE
            self.my_matcher = equal_to_ignoring_case

        elif matcher_type == MatcherType.EQUAL_TO_IGNORE_WHITESPACE:
            self.matcher_type = MatcherType.EQUAL_TO_IGNORE_WHITESPACE
            self.my_matcher = equal_to_ignoring_whitespace

        else:
            raise NotImplementedError(f"Matcher for {matcher_type} not implemented")

    def is_match(self, match_values, data_record) -> bool:

        if not self.validate_key_exists(data_record):
            return False

        if len(match_values) == 0:
            self.my_logger.warning("No Match Values provided, raising Error")
            raise NotImplementedError(f"Cannot use {self.matcher_type.value} to check for "
                                      "empty string. Use None or Not_None.")

        elif isinstance(match_values, list):

            if len(match_values) == 1:
                q_match_values = pull_val(*match_values)
                return (match_equality(self.my_matcher(q_match_values))
                        == str(data_record[self.match_col_key]))
            else:
                matches_list = [q for q in match_values
                                if match_equality(self.my_matcher(q))
                                == str(data_record[self.match_col_key])]
                if len(matches_list) > 0:
                    return True
                else:
                    return False
        else:
            return (match_equality(self.my_matcher(match_values))
                    == str(data_record[self.match_col_key]))


class NumberComparer(Matcher):

    def __init__(self, match_col_key, matcher_type, convert_none_to=None):
        super().__init__(match_col_key)
        self.match_col_key = match_col_key
        self.replacement_val = None
        if convert_none_to is None:
            self.convert_none = False
        else:
            self.convert_none = True
            self.replacement_val = convert_none_to

        if matcher_type == MatcherType.GREATER_THAN:
            self.matcher_type = MatcherType.GREATER_THAN
            self.my_matcher = greater_than

        elif matcher_type == MatcherType.GREATER_THAN_EQUAL_TO:
            self.matcher_type = MatcherType.GREATER_THAN_EQUAL_TO
            self.my_matcher = greater_than_or_equal_to

        elif matcher_type == MatcherType.LESS_THAN:
            self.matcher_type = MatcherType.LESS_THAN
            self.my_matcher = less_than

        elif matcher_type == MatcherType.LESS_THAN_EQUAL_TO:
            self.matcher_type = MatcherType.LESS_THAN_EQUAL_TO
            self.my_matcher = less_than_or_equal_to

        elif matcher_type == MatcherType.CLOSE_TO:
            self.matcher_type = MatcherType.CLOSE_TO
            self.my_matcher = close_to
        else:
            raise NotImplementedError(f"Matcher for {matcher_type} not implemented")

    def get_record_value(self, data_record) -> int:
        """  If you want to convert a None to an Int, set a replacement value. """
        if self.convert_none and data_record[self.match_col_key] is None:
            return_val = self.replacement_val
        else:
            return_val = data_record[self.match_col_key]
        return return_val

    def is_match(self, match_values, data_record) -> bool:
        if not self.validate_key_exists(data_record):
            return False

        if isinstance(match_values, list):
            if len(match_values) == 0:
                self.my_logger.warning("No Match Values provided, raising Error")
                cls_name = self.__class__.__name__
                raise NotImplementedError(fr"Cannot use {cls_name} to check for "
                                          "empty string. Use None or Not_None.")
            elif len(match_values) == 1:
                q_match_values = pull_val(*match_values)
                test_val = self.get_record_value(data_record)
                return (match_equality(self.my_matcher(q_match_values))
                        == test_val)
            else:
                cls_name = self.__class__.__name__
                raise NotImplementedError(fr"Cannot use {cls_name} to check "
                                          " a list of values")
        else:

            if self.matcher_type == MatcherType.CLOSE_TO:
                # Expect a tuple for Close To for Num, Delta...so unpack values
                test_val = self.get_record_value(data_record)
                return (match_equality(self.my_matcher(*match_values))
                        == test_val)
            else:
                test_val = self.get_record_value(data_record)
                return (match_equality(self.my_matcher(match_values))
                        == test_val)


class ExistsMatchers(Matcher):

    def __init__(self, match_col_key, matcher_type):
        super().__init__(match_col_key)
        self.match_col_key = match_col_key

        if matcher_type in (MatcherType.NONE, MatcherType.NONE_OR_EMPTY):
            self.matcher_type = MatcherType(matcher_type)
            self.my_matcher = none

        elif matcher_type in (MatcherType.NOT_NONE, MatcherType.NOT_NONE_OR_EMPTY):
            self.matcher_type = MatcherType(matcher_type)
            self.my_matcher = not_none

        else:
            raise NotImplementedError(f"Matcher for {matcher_type} not implemented")

    def is_match(self, data_record) -> bool:

        if not self.validate_key_exists(data_record):
            return False

        if self.matcher_type in (MatcherType.NONE, MatcherType.NOT_NONE):
            return (match_equality(self.my_matcher())
                    == data_record[self.match_col_key])
        else:
            r""" Special case of chaining equal to with none """
            if self.matcher_type == MatcherType.NONE_OR_EMPTY:
                result = match_equality(any_of(
                    self.my_matcher(),
                    equal_to(''), equal_to([]), equal_to({}), equal_to(())
                )) == data_record[self.match_col_key]
                return result

            elif self.matcher_type == MatcherType.NOT_NONE_OR_EMPTY:
                result = match_equality(all_of(
                    self.my_matcher(),
                    is_not(equal_to('')), is_not(equal_to([])),
                    is_not(equal_to({})), is_not(equal_to(()))
                )) == data_record[self.match_col_key]
                return result
            else:
                raise NotImplementedError(f"Matcher for {self.matcher_type} not implemented")


class DictMatchers(Matcher):

    def __init__(self, match_col_key, matcher_type):
        super().__init__(match_col_key)
        self.match_col_key = match_col_key

        if matcher_type == MatcherType.HAS_ENTRY:
            self.matcher_type = MatcherType(matcher_type)
            self.my_matcher = has_entry

        elif matcher_type == MatcherType.HAS_ENTRIES:
            self.matcher_type = MatcherType(matcher_type)
            self.my_matcher = has_entries

        else:
            raise NotImplementedError(f"Matcher for {matcher_type} not implemented")

    def is_match(self, *match_values, data_record) -> bool:
        if not self.validate_key_exists(data_record):
            return False

        if not data_record[self.match_col_key]:
            """ If record is None or empty, no match"""
            return False

        if self.matcher_type == MatcherType.HAS_ENTRY and len(match_values) > 2:
            raise ValueError("HAS_ENTRY matcher only accepts two values")

        if isinstance(match_values[0], list) and isinstance(data_record[self.match_col_key], list):
            """ Two lists... """
            check_list = []
            for m_dict in match_values[0]:
                # Unpack list of dictionaries to list of tuples to list of match_values
                mv = [item for tup in m_dict.items() for item in tup]
                check_list.append(match_equality(has_item(has_entries(*mv))) == list(data_record[self.match_col_key]))
            return any(check_list)

        if isinstance(match_values[0], list) and isinstance(data_record[self.match_col_key], dict):
            """ Two lists... """
            check_list = []
            for m_dict in match_values[0]:
                # Unpack list of dictionaries to list of tuples to list of match_values
                mv = [item for tup in m_dict.items() for item in tup]
                check_list.append(match_equality(self.my_matcher(*mv)) == data_record[self.match_col_key])
            return any(check_list)

        if isinstance(data_record[self.match_col_key], list):
            # has_item will iterate a sequence ...
            return match_equality(has_item(has_entry(*match_values))) == list(data_record[self.match_col_key])

        if isinstance(match_values[0], dict):
            check_list = []
            for m_dict in match_values:
                # Unpack list of dictionaries to list of tuples to list of match_values
                mv = [item for tup in m_dict.items() for item in tup]
                check_list.append(match_equality(self.my_matcher(*mv)) == data_record[self.match_col_key])
            return any(check_list)

        return (match_equality(self.my_matcher(*match_values))
                == data_record[self.match_col_key])



from unittest import TestCase
from hamcrest import *
from deftlariat import AnythingMatcher, EqualTo, \
    Matcher, MatcherType, NumberComparer, TextComparer, ExistsMatchers, DictMatchers


class TestAnythingMatcher(TestCase):
    def test_dumb(self):
        assert_that(True, equal_to(True), "Dumb test.")

    def test_record_is_match(self):
        a_matcher = AnythingMatcher("full_record")
        test_data = dict()
        assert_that(a_matcher.is_match('full_record', test_data), equal_to(True))

        test_obj = object()
        assert_that(a_matcher.is_match('full_record', test_obj), equal_to(True))

        assert_that(a_matcher.is_match('full_record', None), equal_to(True))

        assert_that(a_matcher.is_match('full_record', 1), equal_to(True))

        assert_that(a_matcher.is_match('fish', 1), equal_to(True))

        assert_that(a_matcher.is_match('dog', 'cat'), equal_to(True))


class TestEqualTo(TestCase):
    def test_is_match_single(self):
        test_data_record = {'name': 'Scarlet Shelton'}
        name_check = EqualTo('name')

        target_value = 'Scarlet Shelton'
        assert_that(name_check.is_match(target_value, test_data_record),
                    equal_to(True),
                    "Is target value in test data?")

        target_value = 'Bob Fisher'
        assert_that(name_check.is_match(target_value, test_data_record),
                    equal_to(False),
                    "Is target value in test data?")

    def test_empty_list_raises_error(self):
        test_data_record = {'name': 'Scarlet Shelton'}

        target_value = []
        name_check = EqualTo('name')

        with self.assertRaises(NotImplementedError):
            name_check.is_match(target_value, test_data_record)

    def test_is_match_list(self):
        test_data_record = {'name': 'Scarlet Shelton'}
        name_check = EqualTo('name')

        # Test a single item list
        target_value = ['Scarlet Shelton']
        self.assertTrue(name_check.is_match(target_value, test_data_record),
                        "Check against single item list")

        # Test successful, but look for one of many people
        target_value = ['Rudy Stout', 'Todd Lee', 'Scarlet Shelton']
        self.assertTrue(name_check.is_match(target_value, test_data_record),
                        "Check against multi item list")

        # Remove Scarlet and test again, should not be found
        target_value.pop(2)
        self.assertFalse(name_check.is_match(target_value, test_data_record), "Check against list")

    def test_is_key_not_in_data(self):
        test_data_record = {'full_name': 'Scarlet Shelton'}

        # Test a single item list
        target_value = ['Scarlet Shelton']
        name_check = EqualTo('name')

        self.assertFalse(name_check.is_match(target_value, test_data_record),
                         "Check against single item list")

    def test_boolean(self):
        test_data_record = {'full_name': True}

        target_value = [False]
        name_check = EqualTo('full_name')

        self.assertFalse(name_check.is_match(target_value, test_data_record),
                         "Check boolean no match")

        self.assertTrue(name_check.is_match([True], test_data_record),
                        "Check boolean does match")


    def test_obj(self):
        class MyObject(object):
            def __init__(self, something):
                self.foo = something

            def __eq__(self, other):
                return self.foo == other.foo

        a = MyObject('cad')
        test_data_record = {'full_name': [a]}
        print(test_data_record)

        b = MyObject(None)
        name_check = EqualTo('full_name')
        target_value = [b]
        print(name_check.is_match(target_value, test_data_record))
        self.assertFalse(name_check.is_match(target_value, test_data_record),
                         "Check boolean no match")

        target_value = [MyObject('cad'), MyObject('aaa')]
        print(name_check.is_match(target_value, test_data_record))








class TestTextCompareBoundaryCases(TestCase):
    def setUp(self) -> None:

        text_matcher_types = [MatcherType.STARTS_WITH, MatcherType.CONTAINS_STRING,
                              MatcherType.CONTAINS_STRING_IN_ORDER,
                              MatcherType.EQUAL_TO_IGNORE_WHITESPACE,
                              MatcherType.EQUAL_TO_IGNORE_CASE]

        self.my_text_comparer_list = list()
        for type in text_matcher_types:
            self.my_text_comparer_list.append(TextComparer('text_col', type))

    def test_empty_input_list(self):
        test_data_record = {'text_col': "some text value"}

        target_value = []
        for comparer in self.my_text_comparer_list:
            target_value = []
            with self.assertRaises(NotImplementedError, msg=f"Checking {comparer}"):
                comparer.is_match(target_value, test_data_record)

    def test_key_not_in_data_record(self):
        test_data_record = {'rating': 'Superduper'}

        # Test a single item list
        target_value = ['Super']
        for comparer in self.my_text_comparer_list:
            target_value = []
            assert_that(comparer.is_match(target_value, test_data_record),
                        equal_to(False), 'Col not present in data recrod returns FALSE')

    def test_bad_matcher_type(self):

        with self.assertRaises(NotImplementedError):
            self.comparer = TextComparer("Foo", MatcherType.GREATER_THAN_EQUAL_TO)


class TestStartsWith(TestCase):
    def test_dumb(self):
        self.assertTrue(True)

    def test_is_match_multiple_values(self):
        test_data_record = {'equipment': 'baseball'}
        ball_starts_with = TextComparer('equipment', MatcherType.STARTS_WITH)

        # Test a single item list
        target_value = ['base']
        self.assertTrue(ball_starts_with.is_match(target_value, test_data_record),
                        "find one of the balls")

        # Test a longer list
        target_value = ['foot', 'soccer', 'base']
        self.assertTrue(ball_starts_with.is_match(target_value, test_data_record),
                        "find one of the balls")

        # Test a longer list, that will fail
        target_value = ['foot', 'soccer', 'basket']
        self.assertFalse(ball_starts_with.is_match(target_value, test_data_record),
                         "No balls to find")

    def test_is_match_single_value(self):
        test_data_record = {'rating': 'Superduper'}

        # Test a single item list
        target_value = 'Super'
        rating_starts_with = TextComparer('rating', MatcherType.STARTS_WITH)

        self.assertTrue(rating_starts_with.is_match(target_value, test_data_record),
                        "looking at field not in data record")

    def test_is_match_convert_to_str(self):
        test_data_record = {'department': 11223344}
        dept_starts_with = TextComparer('department', MatcherType.STARTS_WITH)

        # Test a single item
        target_value = '1122'
        self.assertTrue(dept_starts_with.is_match(target_value, test_data_record),
                        "find one of departments")

        # Test a single item list
        target_value = ['1122']
        self.assertTrue(dept_starts_with.is_match(target_value, test_data_record),
                        "find one of departments")

        # Test a multiple item list
        target_value = ['2222', '1122', '3333']
        self.assertTrue(dept_starts_with.is_match(target_value, test_data_record),
                        "find one of departments")

        # Test a single item  -- No Match
        target_value = '122'
        self.assertFalse(dept_starts_with.is_match(target_value, test_data_record),
                         "No-Match - one of departments")


class TestEqualToIgnoreCase(TestCase):
    def test_dumb(self):
        self.assertTrue(True)

    def test_is_match_multiple_values(self):
        test_data_record = {'equipment': 'basEBAll'}
        eq_ignore_case = TextComparer('equipment', MatcherType.EQUAL_TO_IGNORE_CASE)

        # Test a single item list
        target_value = ['baseball']
        self.assertTrue(eq_ignore_case.is_match(target_value, test_data_record),
                        "find one of the balls")

        # Test a longer list
        target_value = ['football', 'baseball']
        self.assertTrue(eq_ignore_case.is_match(target_value, test_data_record),
                        "find one of the balls")

        # Test a longer list, that will fail
        target_value = ['foot', 'soccer', 'basket']
        self.assertFalse(eq_ignore_case.is_match(target_value, test_data_record),
                         "No balls to find")

    def test_ignore_whitespace(self):
        test_data_record = {'department': 11223344}
        eq_ignore_whitespace = TextComparer('department', MatcherType.EQUAL_TO_IGNORE_WHITESPACE)

        # Test a single item
        target_value = '11223344'
        self.assertTrue(eq_ignore_whitespace.is_match(target_value, test_data_record),
                        "find one of departments")

        # Test a single item list
        target_value = ['11223344']
        self.assertTrue(eq_ignore_whitespace.is_match(target_value, test_data_record),
                        "find one of departments")

        # Test a multiple item list
        target_value = ['2222', '1122', '11223344']
        self.assertTrue(eq_ignore_whitespace.is_match(target_value, test_data_record),
                        "find one of departments")

        # Test a single item  -- No Match
        target_value = '122'
        self.assertFalse(eq_ignore_whitespace.is_match(target_value, test_data_record),
                         "No-Match - one of departments")

        test_data_record = {'department': '  11223344  '}
        eq_ignore_whitespace = TextComparer('department', MatcherType.EQUAL_TO_IGNORE_WHITESPACE)

        # Test a single item
        target_value = '11223344'
        self.assertTrue(eq_ignore_whitespace.is_match(target_value, test_data_record),
                        "find one of departments")

        # Test a single item list
        target_value = ['11223344']
        self.assertTrue(eq_ignore_whitespace.is_match(target_value, test_data_record),
                        "find one of departments")

        # Test a multiple item list
        target_value = ['2222', '1122', '11223344']
        self.assertTrue(eq_ignore_whitespace.is_match(target_value, test_data_record),
                        "find one of departments")

        # Test a single item  -- No Match
        target_value = '122'
        self.assertFalse(eq_ignore_whitespace.is_match(target_value, test_data_record),
                         "No-Match - one of departments")


class TestNumberComparisons(TestCase):

    def setUp(self):
        self.number_gt_check = NumberComparer('my_number', MatcherType.GREATER_THAN)
        self.number_gte_check = NumberComparer('my_number', MatcherType.GREATER_THAN_EQUAL_TO)
        self.number_lt_check = NumberComparer('my_number', MatcherType.LESS_THAN)
        self.number_lte_check = NumberComparer('my_number', MatcherType.LESS_THAN_EQUAL_TO)
        self.number_close_check = NumberComparer('my_number', MatcherType.CLOSE_TO)

    def test_dumb(self):
        self.assertTrue(True)

    def test_single_value(self):
        test_data_record = {'my_number': 5}

        # Test a single item
        target_value = 4
        self.assertTrue(self.number_gt_check.is_match(target_value, test_data_record),
                        " 5 > 4 = True")
        self.assertTrue(self.number_gte_check.is_match(target_value, test_data_record),
                        " 5 >= True ")
        self.assertFalse(self.number_lt_check.is_match(target_value, test_data_record),
                         " 5 < 4 = False")
        self.assertFalse(self.number_lte_check.is_match(target_value, test_data_record),
                         " 5 <= 4 = False ")
        self.assertTrue(self.number_close_check.is_match((target_value, 2), test_data_record),
                        " 5 close to (4,2) = True")
        self.assertTrue(self.number_close_check.is_match((target_value, 1), test_data_record),
                        " 5 close to (4,2) = True")

        # Test a single item
        target_value = 10

        self.assertFalse(self.number_gt_check.is_match(target_value, test_data_record),
                         " 5 > 10 = False")
        self.assertFalse(self.number_gte_check.is_match(target_value, test_data_record),
                         " 5 >= 10 False ")
        self.assertTrue(self.number_lt_check.is_match(target_value, test_data_record),
                        " 5 < 10 = True")
        self.assertTrue(self.number_lte_check.is_match(target_value, test_data_record),
                        " 5 <= 10 = True ")
        self.assertFalse(self.number_close_check.is_match((target_value, 2), test_data_record),
                         " 5 close to (10,2) = False")
        self.assertFalse(self.number_close_check.is_match((target_value, 1), test_data_record),
                         " 5 close to (10,2) = False")


class TestNumberComparisonsWithReplacement(TestCase):

    def setUp(self):
        self.number_gt_check = NumberComparer('my_number', MatcherType.GREATER_THAN, convert_none_to=5)
        self.number_gte_check = NumberComparer('my_number', MatcherType.GREATER_THAN_EQUAL_TO, convert_none_to=5)
        self.number_lt_check = NumberComparer('my_number', MatcherType.LESS_THAN, convert_none_to=5)
        self.number_lte_check = NumberComparer('my_number', MatcherType.LESS_THAN_EQUAL_TO, convert_none_to=5)
        self.number_close_check = NumberComparer('my_number', MatcherType.CLOSE_TO, convert_none_to=5)

    def test_single_none_with_replacement(self):
        test_data_record = {'my_number': None}

        # Test a single item
        target_value = 4
        self.assertTrue(self.number_gt_check.is_match(target_value, test_data_record),
                        " 5 > 4 = True")
        self.assertTrue(self.number_gte_check.is_match(target_value, test_data_record),
                        " 5 >= True ")
        self.assertFalse(self.number_lt_check.is_match(target_value, test_data_record),
                         " 5 < 4 = False")
        self.assertFalse(self.number_lte_check.is_match(target_value, test_data_record),
                         " 5 <= 4 = False ")
        self.assertTrue(self.number_close_check.is_match((target_value, 2), test_data_record),
                        " 5 close to (4,2) = True")
        self.assertTrue(self.number_close_check.is_match((target_value, 1), test_data_record),
                        " 5 close to (4,2) = True")

        # Test a single item
        target_value = 10

        self.assertFalse(self.number_gt_check.is_match(target_value, test_data_record),
                         " 5 > 10 = False")
        self.assertFalse(self.number_gte_check.is_match(target_value, test_data_record),
                         " 5 >= 10 False ")
        self.assertTrue(self.number_lt_check.is_match(target_value, test_data_record),
                        " 5 < 10 = True")
        self.assertTrue(self.number_lte_check.is_match(target_value, test_data_record),
                        " 5 <= 10 = True ")
        self.assertFalse(self.number_close_check.is_match((target_value, 2), test_data_record),
                         " 5 close to (10,2) = False")
        self.assertFalse(self.number_close_check.is_match((target_value, 1), test_data_record),
                         " 5 close to (10,2) = False")

    def test_is_gt_list_value(self):
        test_data_record = {'my_number': 5}

        # Test a single item list
        target_value = [4]
        self.assertTrue(self.number_gt_check.is_match(target_value, test_data_record),
                        "looking at field not in data record")

        # Test a single item list
        target_value = [10]
        self.assertFalse(self.number_gt_check.is_match(target_value, test_data_record),
                         "looking at field not in data record")

    def test_empty_list_raise_error(self):
        test_data_record = {'my_number': 11223344}

        target_value = []
        with self.assertRaises(NotImplementedError):
            self.number_gt_check.is_match(target_value, test_data_record)

    def test_multiple_vals_raise_error(self):
        test_data_record = {'my_number': 11223344}

        target_value = [1, 2, 3, 4]
        with self.assertRaises(NotImplementedError):
            self.number_gt_check.is_match(target_value, test_data_record)


class AbstractTestMatcher(Matcher):
    """ A class sole for testing the Abstract Base Class """

    def __init__(self, match_key_col):
        super().__init__(match_key_col)

    def is_match(self, *args):
        raise NotImplementedError


class TestMatcher(TestCase):
    def test_reper(self):
        abc_matcher = AbstractTestMatcher("some_key_field")
        assert_that("some_key_field", is_in(abc_matcher.__repr__()), "Check output")

    def test_get_key_value(self):
        abc_matcher = AbstractTestMatcher("some_key_field")
        print(abc_matcher.get_key_val())

        test_key = frozenset(('some_key_field', MatcherType.NOTHING.value))
        assert_that(abc_matcher.get_key_val(), equal_to(test_key))

    def test_str(self):
        abc_matcher = AbstractTestMatcher("some_key_field")
        assert_that("some_key_field", is_in(abc_matcher.__str__()), "Check output")
        print(abc_matcher.__str__())

    def test_match(self):
        abc_matcher = AbstractTestMatcher("some_key_field")
        with self.assertRaises(NotImplementedError):
            abc_matcher.is_match("target", "data_record")


class TestNumberMatcherBoundaryCases(TestCase):
    def setUp(self) -> None:

        number_matcher_types = [MatcherType.GREATER_THAN, MatcherType.GREATER_THAN_EQUAL_TO,
                                MatcherType.LESS_THAN, MatcherType.LESS_THAN_EQUAL_TO,
                                MatcherType.CLOSE_TO]

        self.my_number_comparere_list = list()
        for m_type in number_matcher_types:
            self.my_number_comparere_list.append(NumberComparer('number_col', m_type))

    def test_empty_input(self):
        test_data_record = {'number_col': 11223344}

        for comparer in self.my_number_comparere_list:
            target_value = []
            with self.assertRaises(NotImplementedError, msg=f"Checking {comparer}"):
                comparer.is_match(target_value, test_data_record)

    def test_key_not_in_data_record(self):
        test_data_record = {'some_other_columns': 11223344}

        for comparer in self.my_number_comparere_list:
            target_value = []

            assert_that(comparer.is_match(target_value, test_data_record),
                        equal_to(False), "Return False when field not available.")

    def test_reject_multiple_inputs(self):
        test_data_record = {'number_col': 11223344}

        for comparer in self.my_number_comparere_list:
            target_value = [1, 2, 3, 4]
            with self.assertRaises(NotImplementedError, msg=f"Checking {comparer}"):
                comparer.is_match(target_value, test_data_record)

    def test_bad_matcher_type(self):
        with self.assertRaises(NotImplementedError):
            self.comparer = NumberComparer("Foo", MatcherType.STARTS_WITH)


class TestExistenceMatcherBoundaryCases(TestCase):
    def setUp(self) -> None:

        exists_matcher_types = [MatcherType.NONE, MatcherType.NONE_OR_EMPTY,
                                MatcherType.NOT_NONE, MatcherType.NOT_NONE_OR_EMPTY]

        self.my_exists_comparer_list = list()
        for m_type in exists_matcher_types:
            self.my_exists_comparer_list.append(ExistsMatchers('number_col', m_type))

    def test_key_not_in_data_record(self):
        test_data_record = {'some_other_columns': 11223344}

        for comparer in self.my_exists_comparer_list:
            target_value = []

            assert_that(comparer.is_match(test_data_record),
                        equal_to(False), "Return False when field not available.")

    def test_bad_matcher_type(self):
        with self.assertRaises(NotImplementedError):
            self.comparer = NumberComparer("Foo", MatcherType.STARTS_WITH)


class TestExistsMatchers(TestCase):

    def test_is_none(self):
        my_is_none = ExistsMatchers('some_col', MatcherType.NONE)

        test_data_record = {'some_col': None}
        assert_that(my_is_none.is_match(test_data_record), equal_to(True),
                    "It is actually None")

        test_data_record = {'some_col': 'data'}
        assert_that(my_is_none.is_match(test_data_record), equal_to(False),
                    "It is actually None")

        test_data_record = {'some_col': ''}
        assert_that(my_is_none.is_match(test_data_record), equal_to(False),
                    "It is actually None")

        test_data_record = {'some_col': []}
        assert_that(my_is_none.is_match(test_data_record), equal_to(False),
                    "It is actually None")

    def test_not_none(self):
        my_not_none = ExistsMatchers('some_col', MatcherType.NOT_NONE)

        test_data_record = {'some_col': None}
        assert_that(my_not_none.is_match(test_data_record), equal_to(False),
                    "It is actually None")

        test_data_record = {'some_col': 'data'}
        assert_that(my_not_none.is_match(test_data_record), equal_to(True),
                    "It is actually None")

        test_data_record = {'some_col': ''}
        assert_that(my_not_none.is_match(test_data_record), equal_to(True),
                    "It is actually None")

        test_data_record = {'some_col': []}
        assert_that(my_not_none.is_match(test_data_record), equal_to(True),
                    "It is actually None")

    def test_is_none_or_empty(self):
        my_is_none = ExistsMatchers('some_col', MatcherType.NONE_OR_EMPTY)

        test_data_record = {'some_col': None}
        assert_that(my_is_none.is_match(test_data_record), equal_to(True),
                    "It is actually None")

        test_data_record = {'some_col': 'data'}
        assert_that(my_is_none.is_match(test_data_record), equal_to(False),
                    "It is actually None")

        test_data_record = {'some_col': ''}
        assert_that(my_is_none.is_match(test_data_record), equal_to(True),
                    "It is actually None")

        test_data_record = {'some_col': []}
        assert_that(my_is_none.is_match(test_data_record), equal_to(True),
                    "It is actually None")

        test_data_record = {'some_col': {}}
        assert_that(my_is_none.is_match(test_data_record), equal_to(True),
                    "It is actually None")

        test_data_record = {'some_col': ()}
        assert_that(my_is_none.is_match(test_data_record), equal_to(True),
                    "It is actually None")

    def test_not_none_or_empty(self):
        my_is_none = ExistsMatchers('some_col', MatcherType.NOT_NONE_OR_EMPTY)

        test_data_record = {'some_col': None}
        assert_that(my_is_none.is_match(test_data_record), equal_to(False),
                    "It is actually None")

        test_data_record = {'some_col': 'data'}
        assert_that(my_is_none.is_match(test_data_record), equal_to(True),
                    "It is actually None")

        test_data_record = {'some_col': ''}
        assert_that(my_is_none.is_match(test_data_record), equal_to(False),
                    "It is actually None")

        test_data_record = {'some_col': []}
        assert_that(my_is_none.is_match(test_data_record), equal_to(False),
                    "It is actually None")

        test_data_record = {'some_col': {}}
        assert_that(my_is_none.is_match(test_data_record), equal_to(False),
                    "It is actually None")

        test_data_record = {'some_col': ()}
        assert_that(my_is_none.is_match(test_data_record), equal_to(False),
                    "It is actually None")


class TestDictMatchers(TestCase):

    def test_has_entries(self):
        operator = DictMatchers('dumb', MatcherType.HAS_ENTRIES)
        assert_that(operator.is_match('a', 'b', data_record={'dumb': {'a': 'b'}}),
                    equal_to(True))

        assert_that(operator.is_match({'a': 'b'}, data_record={'dumb': {'a': 'b'}}),
                    equal_to(True))

        assert_that(operator.is_match({'foo': 'bar'}, data_record={'dumb': {'a': 'b'}}),
                    equal_to(False))
        assert_that(operator.is_match('foo', 'bar', data_record={'dumb': {'a': 'b'}}),
                    equal_to(False))

        assert_that(operator.is_match('a', 'b', 'c', 'd',
                                      data_record={'dumb': {'c': 'd', 'a': 'b'}}),
                    equal_to(True))

        assert_that(operator.is_match('a', 'b',
                                      data_record={'dumb': [
                                          {'1': '2'}, {'c': 'd'}, {'a': 'b'}]}),
                    equal_to(True))

        assert_that(operator.is_match([{'a': 'b'}, {'1': '2'}],
                                      data_record={'dumb': [
                                          {'1': '2'}, {'c': 'd'}, {'a': 'b'}]}),
                    equal_to(True))

        assert_that(operator.is_match([{'a': 'g'}, {'1': 'g'}],
                                      data_record={'dumb': [
                                          {'1': '2'}, {'c': 'd'}, {'a': 'b'}]}),
                    equal_to(False))

        assert_that(operator.is_match([{'a': 'g'}, {'1': 'g'}],
                                      data_record={'dumb': {'a': 'b'}}),
                    equal_to(False))

        assert_that(operator.is_match([{'a': 'b'}, {'1': '2'}],
                                      data_record={'dumb': {'a': 'b'}}),
                    equal_to(True))

        assert_that(operator.is_match([{'a': 'b'}, {'1': '2'}, {'id': 'asdf', 'identifier': 'foo'}],
                                      data_record={'dumb': [{'a': 'b'}, {'id': 'asdf', 'identifier': 'foo'}]}),
                    equal_to(True))

        assert_that(operator.is_match([{'a': 'b'}, {'1': '2'}],
                                      data_record={'dumb': None}),
                    equal_to(False))



    def test_has_entry(self):
        operator = DictMatchers('dumb', MatcherType.HAS_ENTRY)
        assert_that(operator.is_match('a', 'b', data_record={'dumb': {'a': 'b'}}),
                    equal_to(True))

        assert_that(operator.is_match({'a': 'b'}, data_record={'dumb': {'a': 'b'}}),
                    equal_to(True))

        assert_that(operator.is_match({'foo': 'bar'}, data_record={'dumb': {'a': 'b'}}),
                    equal_to(False))
        assert_that(operator.is_match('foo', 'bar', data_record={'dumb': {'a': 'b'}}),
                    equal_to(False))

        assert_that(operator.is_match('a', 'b',
                                      data_record={'dumb': [
                                          {'1': '2'}, {'c': 'd'}, {'a': 'b'}]}),
                    equal_to(True))

        assert_that(operator.is_match([{'a': 'b'}, {'1': '2'}],
                                      data_record={'dumb': [
                                          {'1': '2'}, {'c': 'd'}, {'a': 'b'}]}),
                    equal_to(True))


        assert_that(operator.is_match([{'a': 'g'}, {'1': 'g'}],
                                      data_record={'dumb': [
                                          {'1': '2'}, {'c': 'd'}, {'a': 'b'}]}),
                    equal_to(False))

        assert_that(operator.is_match([{'a': 'g'}, {'1': 'g'}],
                                      data_record={'dumb': {'a': 'b'}}),
                    equal_to(False))

        assert_that(operator.is_match([{'a': 'b'}, {'1': '2'}],
                                      data_record={'dumb': {'a': 'b'}}),
                    equal_to(True))

        assert_that(operator.is_match([{'a': 'b'}, {'1': '2'}],
                                      data_record={'dumb': None}),
                    equal_to(False))



    def test_has_entry_bad_input(self):
        operator = DictMatchers('dumb', MatcherType.HAS_ENTRY)

        with self.assertRaises(ValueError):
            assert_that(operator.is_match('a', 'b', 'c', 'd',
                                          data_record={'dumb': {'c': 'd', 'a': 'b'}}),
                        equal_to(True))

    def test_this(self):
        target_skills = [{'id': '61d8923376841a071b5fb398', 'identifier': 'Alto Sax'}, {'id': '61d8923376841a071b5fb39a', 'identifier': 'Tenor Sax'}, {'id': '61d8923376841a071b5fb39c', 'identifier': 'Bari Sax'}, {'id': '61d8923376841a071b5fb39e', 'identifier': 'Trumpet'}, {'id': '61d8923376841a071b5fb3a0', 'identifier': 'Trombone'}, {'id': '6223aa9d10a7d3001e006f66', 'identifier': '2nd Trumpet'}]





"""Console script for src."""
import sys
import json

import click

from deftlariat import __version__
from deftlariat import EqualTo, NumberComparer, MatcherType


@click.group(no_args_is_help=True)
@click.version_option(version=__version__)
def deft_cli():
    """Console script for deft lariats."""
    pass


@deft_cli.command()
def hello(name: str) -> None:
    click.echo(f"Hello {name}")


zip_code_list = ['']


@deft_cli.command()
@click.option('--data-file',
              help='file to work with',
              type=click.File('r'),
              default=sys.stdin)
def example_coingecko(data_file) -> None:
    """
    Coin Gecko Example. Demonstrate how to use the data filter with a list of dictionaries from
    coingecko.com API. Return records that match any one of three configured filters.
    """
    click.echo("Example One - Data Filter.  \nThis example will search the input list of dictionaries for "
               "records that match any of three filters.\n\n"
               "The three filters are:\n"
               "1. EqualTo - Check if the value of a the `symbol` field is equal to a target value `OTCMKTS:FRMO`.\n"
               "2. NumberComparer - Check if the value of the `total_holdings` field is greater than or equal to "
               " the target value: `1000`\n"
               "3. NumberComparer - Check if the value of the `percentage_of_total_supply` field is greater than "
               "or equal to the target value: `0.1`\n"
               "Matching records will be printed to stdout.\n\n\n"
               )

    with data_file as f:
        try:
            # First read the data as a string
            content = f.read()

            # Check if there's any content
            if not content.strip():
                # No Data on stdin, so return
                click.echo("No Data found - Either File or stdin")
                return

            # Then parse it as JSON
            data = json.loads(content)
            # If data is not a list, wrap it in a list
            if not isinstance(data, list):
                data = [data]
        except json.JSONDecodeError as e:
            click.echo(f"Error parsing JSON: {e}")
            return

    # TODO: Assume stream is a json list
    field_key = 'symbol'
    filter_one = EqualTo(field_key)
    target_value = 'OTCMKTS:FRMO'

    filter_two = NumberComparer('total_holdings', MatcherType.GREATER_THAN_EQUAL_TO)
    filter_three = NumberComparer('percentage_of_total_supply', MatcherType.GREATER_THAN_EQUAL_TO)
    for x in data:
        if any([filter_one.is_match(target_value, x),
                filter_two.is_match(1000, x),
                filter_three.is_match(0.1, x),
                ]):
            click.echo(f"Data Filter hit for record:\n{x}\n\n")



#!/usr/bin/env python

"""Tests for `src` package."""


import unittest

from click.testing import CliRunner
from deftlariat.scripts import cli
from hamcrest import *


class TestSrc(unittest.TestCase):
    """Tests for `src` package."""

    def setUp(self):
        """Set up test fixtures, if any."""

    def tearDown(self):
        """Tear down test fixtures, if any."""

    def test_000_something(self):
        """Test something."""
        pass

    def test_command_line_interface(self):
        """Test the CLI."""
        runner = CliRunner()
        result = runner.invoke(cli.deft_cli)
        print(result.output)
        assert_that(result.exit_code, equal_to(2))  # Should be SystemExit issue, no command
        assert 'deft lariats' in result.output
        help_result = runner.invoke(cli.deft_cli, ['--help'])
        assert_that( help_result.exit_code, equal_to(0))
        assert_that(help_result.output, string_contains_in_order('--version', '--help'))

    def test_example_one_with_json_input(self):
        """Test the example_one command with JSON input."""
        runner = CliRunner()

        # Test with a single JSON object using a temporary file
        with runner.isolated_filesystem():
            with open('test.json', 'w') as f:
                f.write('[{"symbol": "OTCMKTS:FRMO", "total_holdings": 2000, "percentage_of_total_supply": 0.2}]')

            result = runner.invoke(cli.deft_cli, ['example-coingecko', '--data-file', 'test.json'], catch_exceptions=False)
            print(f"Single JSON output: {result.output}")
            print(f"Exit code: {result.exit_code}")
            if result.exception:
                print(f"Exception: {result.exception}")
            assert result.exit_code == 0
            assert 'Data Filter hit' in result.output

        # Test with a JSON array using a temporary file
        with runner.isolated_filesystem():
            with open('test.json', 'w') as f:
                f.write('[{"symbol": "OTCMKTS:FRMO", "total_holdings": 2000, "percentage_of_total_supply": 0.2}]')

            result = runner.invoke(cli.deft_cli, ['example-coingecko', '--data-file', 'test.json'], catch_exceptions=False)
            print(f"JSON array output: {result.output}")
            print(f"JSON array exit code: {result.exit_code}")
            if result.exception:
                print(f"JSON array exception: {result.exception}")
            assert result.exit_code == 0
            assert 'Data Filter hit' in result.output

        # Test with JSON that doesn't match the filter criteria
        with runner.isolated_filesystem():
            with open('test.json', 'w') as f:
                f.write('[{"symbol": "BTC", "total_holdings": 10, "percentage_of_total_supply": 0.01}]')

            result = runner.invoke(cli.deft_cli, ['example-coingecko', '--data-file', 'test.json'], catch_exceptions=False)
            print(f"Non-matching JSON output: {result.output}")
            print(f"Non-matching JSON exit code: {result.exit_code}")
            if result.exception:
                print(f"Non-matching JSON exception: {result.exception}")
            assert result.exit_code == 0
            # No filter hit, so no output expected
            assert 'Data Filter hit' not in result.output



#!/usr/bin/env python

"""The setup script."""

import pathlib
from os import path

from setuptools import setup, find_namespace_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

# automatically captured required modules for install_requires in requirements.txt
# and as well as configure dependency links
HERE = pathlib.Path(__file__).parent
with open(path.join(HERE, 'requirements.txt'), encoding='utf-8') as f:
    all_reqs = f.read().split('\n')
install_requirements = [x.strip() for x in all_reqs if ('git+' not in x) and (
    not x.startswith('#')) and (not x.startswith('-'))]
dependency_links = [x.strip().replace('git+', '') for x in all_reqs \
                    if 'git+' not in x]


test_requirements = []

setup(
    name='PyDeftLariats',
    author="Tyler McMaster",
    author_email='mcmasty@yahoo.com',
    python_requires='>=3.11',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    description="Using PyHamcrest to build a collection of data filters.",
    entry_points={
        'console_scripts': [
            'deft=deftlariat.scripts.cli:deft_cli',
        ],
    },
    install_requires=install_requirements,
    license="GNU General Public License v3",
    keywords="hamcrest matchers data filters",
    long_description=readme + '\n\n' + history,
    long_description_content_type='text/x-rst',
    include_package_data=True,
    packages=find_namespace_packages("src"),
    package_dir={"": "src"},
    package_data={"deftlariat": ["py.typed"]},
    provides=['deftlariat'],
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/mcmasty/PyDeftLariats',
    version='1.2.11',
    zip_safe=False,
)
